// Copyright (c) 2009-2015 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly


#include "host_common.h"

//basic components of any cl sample application
cl_context g_clContext;
cl_device_id *g_clDevices = NULL;
cl_command_queue g_clCommandQueue;
cl_program g_clProgram;
char *g_clProgramString;


//data structures specific to this sample solution
float *g_f32_resultImage = NULL; //fimg

unsigned char *g_img = NULL;
unsigned int g_h,g_w;
//unsigned int g_numSubSamples = 1; //for speed, go low, accuracy, go high
unsigned int g_imageSize;
bool g_bAlignedAlloc = ALIGNED_ALLOCATION;

//for generic addy space
cl_mem g_cl_mem_inputImage;
unsigned char inX[IMAGE_HEIGHT * IMAGE_WIDTH];
cl_mem g_cl_mem_resultImage;

//specific for generic addy space
cl_kernel cl_kernel_GlobalXToY;
cl_kernel cl_kernel_LocalXToY;
//cl_kernel cl_kernel_GenericXToY;




//if an error occurs we exit
//it would be better to cleanup state then exit, for sake of simplicity going to omit the cleanup
void testStatus(int status, char *errorMsg)
{
	if(status != SUCCESS)
	{
		if(errorMsg == NULL)
		{
			printf("Error\n");
		}
		else
		{
			printf("Error: %s", errorMsg);
		}
		exit(EXIT_FAILURE);
	}
}

int HandleCompileError(void)
{
	cl_int logStatus;
	char *buildLog = NULL;
	size_t buildLogSize = 0;
	//in this tutorial i only have one device
	logStatus = clGetProgramBuildInfo( g_clProgram, g_clDevices[0], CL_PROGRAM_BUILD_LOG, buildLogSize, buildLog, &buildLogSize);
	if(logStatus != CL_SUCCESS)
	{
		printf("logging error\n");
		exit(EXIT_FAILURE);
	}

	buildLog = (char *)malloc(buildLogSize);
	if(buildLog == NULL)
	{
		printf("ERROR TO ALLOCATE MEM FOR BUILDLOG\n");
		exit(EXIT_FAILURE);
	}

	memset(buildLog, 0, buildLogSize);

	logStatus = clGetProgramBuildInfo (g_clProgram, g_clDevices[0], CL_PROGRAM_BUILD_LOG, buildLogSize, buildLog, NULL);
	if(logStatus != CL_SUCCESS)
	{
		free(buildLog);
		return FAIL;
	}

	printf("\nBUILD LOG\n");
	printf("************************************************************************\n");
	printf("%s END \n", buildLog);
	printf("************************************************************************\n");
	free(buildLog);
	return SUCCESS;
}


//convert to string needs to take a string a file as input and write a char * to output
int convertToStringBuf(const char *fileName)
{
	FILE *fp = NULL;
	int status;

	fopen_s(&fp, fileName, "r");
	if(fp == NULL)
	{
		printf("Error opening %s, check path\n", fileName);
		exit(EXIT_FAILURE);
	}

	status = fseek(fp, 0, SEEK_END);
	if(status != 0)
	{
		printf("Error finding end of file\n");
		exit(EXIT_FAILURE);
	}

	int len = ftell(fp);
	if(len == -1L)
	{
		printf("Error reporting position of file pointer\n");
		exit(EXIT_FAILURE);
	}
	rewind(fp);
	g_clProgramString = (char *)malloc((len * sizeof(char))+1);
	if(g_clProgramString == NULL)
	{
		printf("Error in allocation when converting CL source file to a string\n");
		exit(EXIT_FAILURE);
	}
	memset(g_clProgramString, '\0', len+1);
	fread(g_clProgramString, sizeof(char), len, fp);
	status = ferror(fp);
	if(status != 0)
	{
		printf("Error reading into the program string from file\n");
		exit(EXIT_FAILURE);
	}
	fclose(fp);
	
	return SUCCESS;
}


int initializeHost(void)
{
	g_clDevices = NULL;
	g_clProgramString = NULL;
	g_f32_resultImage = NULL;
	g_img = NULL;
	
	return SUCCESS;
}
 

int initializeDeviceCode()
{
	cl_int status;

	//load CL file, build CL program object, create CL kernel object
	const char *filename = CL_KERNELS;
	status = convertToStringBuf(filename);

	size_t sourceSize = strlen(g_clProgramString);

	g_clProgram = clCreateProgramWithSource(g_clContext, 1, (const char **)&g_clProgramString, &sourceSize, &status);
	testStatus(status, "clCreateProgramWithSource error");

//	status = clBuildProgram(g_clProgram, 1, g_clDevices, NULL, NULL, NULL);
	status = clBuildProgram(g_clProgram, 1, g_clDevices, "-cl-std=CL2.0", NULL, NULL);
	if(status != CL_SUCCESS)
	{
		if(status == CL_BUILD_PROGRAM_FAILURE)
		{
			HandleCompileError();
		} //end if BUILD_PROGRAM_FAILURE
	} //end if CL_SUCCESS

	cl_kernel_GlobalXToY = clCreateKernel(g_clProgram, "GlobalXToY_OpenCL2_0", &status);
	testStatus(status, "clCreatekernel error");
	cl_kernel_LocalXToY = clCreateKernel(g_clProgram, "LocalXToY_OpenCL_2_0", &status);
	testStatus(status, "clCreatekernel error");
	
	return SUCCESS;

}

int initializeCL()
{
	cl_int status = 0;
	cl_uint numPlatforms= 0;

	//[1] get the platform
	cl_platform_id platformToUse = NULL;
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	testStatus(status, "clGetPlatformIDs error");
	
	cl_platform_id *platforms = (cl_platform_id *)malloc(sizeof(cl_platform_id) * numPlatforms);
	if(platforms == NULL)
	{
		printf("Error when allocating space for the platforms\n");
		exit(EXIT_FAILURE);
	}

	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	testStatus(status, "clGetPlatformIDs error");

	for(unsigned int i=0;i<numPlatforms;i++)
	{
		char pbuf[100];
		status = clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, sizeof(pbuf), pbuf, NULL);
		testStatus(status, "clGetPlatformInfo error");

		if(strstr(pbuf, "OpenCL 2") != NULL)
		{
			printf("Great! We found an OpenCL 2.0 platform.\n");
			platformToUse = platforms[i];
			break;
		}
	}
		
	if(platformToUse == NULL)
	{
		printf("We have not found an Intel(r) OpenCL implementation, exiting application\n");
		exit(EXIT_FAILURE);
	}

	//[2] get device ids for the platform i have obtained
	cl_uint num_devices = -1; //yeah yeah, i know uint 
	cl_device_info devTypeToUse = CL_DEVICE_TYPE_DEFAULT;  //set as CPU or GPU from host_common.h

	//get # of devices of this type on this platform and allocate space in g_clDevices (better be 1 for this tutorial)
	status = clGetDeviceIDs(platformToUse, devTypeToUse, 0, g_clDevices, &num_devices);
	testStatus(status, "clGetDeviceIDs error, might need to set value CL_DEVICE_TYPE_CPU_OR_GPU");
	//allocate space
	g_clDevices = (cl_device_id *)malloc(sizeof(cl_device_id)*num_devices);
	if(g_clDevices == NULL)
	{
		printf("Error when creating space for devices\n");
		exit(EXIT_FAILURE);
	}
	//we know we have an intel platform, get the device we want to use
	status = clGetDeviceIDs(platformToUse, devTypeToUse, num_devices, g_clDevices, 0);
	testStatus(status, "clGetDeviceIDs error");

	//print out the device type just to make sure we got it right
	cl_device_type device_type;
    char vendorName[255];
	memset(vendorName, '\0', 255);

	clGetDeviceInfo(g_clDevices[0], CL_DEVICE_TYPE, sizeof(cl_device_type), (void *)&device_type, NULL);

	clGetDeviceInfo(g_clDevices[0], CL_DEVICE_VENDOR, (sizeof(char)*256), vendorName, NULL);


	if(device_type == CL_DEVICE_TYPE_CPU)
	{
		printf("Device type is CPU, Vendor is %s\n", vendorName);
	}
	else if(device_type == CL_DEVICE_TYPE_GPU)
	{
		printf("Device type is GPU, Vendor is %s\n", vendorName);
	}
	else 
	{
		printf("device type is unknown\n");
	}

	cl_context_properties cps[3] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platformToUse, 0 };

	//create an OCL context
	g_clContext = clCreateContext(cps, 1, g_clDevices, NULL, NULL, &status);
	testStatus(status, "clCreateContext error");

	//create an openCL commandqueue
	g_clCommandQueue = clCreateCommandQueueWithProperties(g_clContext, g_clDevices[0], 0, &status);
	testStatus(status, "clCreateCommandQueue error");
	
	//create device side program, compile and create program objects
	status = initializeDeviceCode();
	testStatus(status, "error in initializeDevice()");

	//create device side buffers
	status = initializeDeviceData();
	testStatus(status, "error in initializeDeviceData()");
	
	if(numPlatforms > 0)
	{
		free(platforms);
	}

	if(g_clDevices != NULL)
	{
		free(g_clDevices);
	}

	return SUCCESS;
}


int cleanupCL()
{
	//cleanup all CL queues, contexts, programs, mem_objs, kernels
	cl_int status;

	status = clReleaseKernel(cl_kernel_GlobalXToY);
	testStatus(status, "Error releasing kernel");

	status = clReleaseKernel(cl_kernel_LocalXToY);
	testStatus(status, "Error releasing kernel");

	status = clReleaseProgram(g_clProgram);
	testStatus(status, "Error releasing program");

	status = clReleaseMemObject(g_cl_mem_inputImage);
	testStatus(status, "Error releasing mem object");

	status = clReleaseMemObject(g_cl_mem_resultImage);
	testStatus(status, "Error releasing mem object");

	status = clReleaseCommandQueue(g_clCommandQueue);
	testStatus(status, "Error releasing mem object");

	status = clReleaseContext(g_clContext);
	testStatus(status, "Error releasing mem object");

	return status;
}


int initializeDeviceData()
{	
	//initiate device side data objects (buffers and values)
	cl_int status = 0;

	//initialize host side data
	g_h = IMAGE_HEIGHT;
	g_w = IMAGE_WIDTH;
	g_bAlignedAlloc = false; //using aligned allocation for the output image?

	g_imageSize = sizeof(unsigned char) * g_w * g_h * NUM_IMAGE_LAYERS;

	//for PPM only: note unsigned char, not float, for ppm file format
	g_img = (unsigned char *)malloc( g_w * g_h * 3);
	if(g_img == NULL)
	{
		printf("Error in initializeDeviceData(), can't allocate space for resulting image\n");
		exit(EXIT_FAILURE);
	}
	memset((void *)g_img, 0, g_w * g_h * 3);

	//allocate the OCL buffer memory objects 
	//notice the NULL here, not passing in a pointer so it is wherever driver wants to put it 
	//intel driver will allocate so it can be zero copy
	
//	initRGBBuf(inRGB);
//	printRGB8Buf(inRGB);

	
	for(unsigned int i=0;i<IMAGE_HEIGHT*IMAGE_WIDTH;i++) 
	{
		inX[i] = (unsigned char)i;
	}


	g_cl_mem_inputImage = clCreateBuffer(g_clContext, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, g_w * g_h, inX, &status);
	testStatus(status, "clCreateBuffer error");

	g_cl_mem_resultImage = clCreateBuffer(g_clContext, CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR, g_w * g_h, NULL, &status);
	testStatus(status, "clCreateBuffer error");

	return status;

}

int runCLKernels(void)
{
	unsigned int status = 0;
	
	//note we have to call from host to device with different signatures but on the device they use the 
	//same internal function
	
	status = RunGlobalOpenCL2_0Version();
	status = RunLocalOpenCL2_0Version();

	return status;
}

int RunGlobalOpenCL2_0Version()
{
	int status;
	unsigned char *mappedBuffer = NULL;

	status = clSetKernelArg(cl_kernel_GlobalXToY, 0, sizeof(cl_mem), (void *)&g_cl_mem_inputImage);
	testStatus(status, "clSetKernel Arg");

	status = clSetKernelArg(cl_kernel_GlobalXToY, 1, sizeof(cl_mem), (void *)&g_cl_mem_resultImage);
	testStatus(status, "clSetKernel Arg");

	int startIn = 0;
	int startOut = 0;
	int length = 1;

	status = clSetKernelArg(cl_kernel_GlobalXToY, 2, sizeof(cl_int), (void *) &startIn);
	testStatus(status, "clSetKernel Arg");
	
	status = clSetKernelArg(cl_kernel_GlobalXToY, 3, sizeof(cl_int), (void *) &startOut);
	testStatus(status, "clSetKernel Arg");

	status = clSetKernelArg(cl_kernel_GlobalXToY, 4, sizeof(cl_int), (void *) &length);
	testStatus(status, "clSetKernel Arg");
	
	//execute the kernel
	size_t global_dim[1];
	global_dim[0] = g_h * g_w;
	size_t local_dim[1];
	local_dim[0] = 4; //local size


	status = clEnqueueNDRangeKernel(g_clCommandQueue, cl_kernel_GlobalXToY, 1, NULL, global_dim, local_dim, 0, NULL, NULL);
	testStatus(status, "clEnqueueNDRangeKernel error");

	clFinish(g_clCommandQueue);

	//read back the results
	//consider putting a flush here, ensures i already put compute aspect
	mappedBuffer = (unsigned char *)clEnqueueMapBuffer(g_clCommandQueue, g_cl_mem_resultImage, CL_TRUE, CL_MAP_READ, 0, g_w * g_h, 0, NULL, NULL, NULL);

	printf("Results from GLOBAL version:\n");
	for(unsigned int i=0;i<g_w * g_h; i++)
	{
		printf("%d is out[%d]\n", mappedBuffer[i], i);
	}
	printf("END\n");

	clEnqueueUnmapMemObject(g_clCommandQueue, g_cl_mem_resultImage, mappedBuffer, 0, NULL, NULL);
	clFinish(g_clCommandQueue);

	return status;

}

int RunLocalOpenCL2_0Version()
{
	int status;
	unsigned char *mappedBuffer = NULL;

	status = clSetKernelArg(cl_kernel_LocalXToY, 0, sizeof(cl_mem), (void *)&g_cl_mem_inputImage);
	testStatus(status, "clSetKernel Arg");

	status = clSetKernelArg(cl_kernel_LocalXToY, 1, sizeof(cl_mem), (void *)&g_cl_mem_resultImage);
	testStatus(status, "clSetKernel Arg");

	int startIn = 0;
	int startOut = 0;

	//execute the kernel
	size_t global_dim[1];
	global_dim[0] = g_h * g_w;
	size_t local_dim[1];
	local_dim[0] = 4; //local size

	status = clEnqueueNDRangeKernel(g_clCommandQueue, cl_kernel_LocalXToY, 1, NULL, global_dim, local_dim, 0, NULL, NULL);
	testStatus(status, "clEnqueueNDRangeKernel error");

	clFinish(g_clCommandQueue);

	//read back the results
	//consider putting a flush here, ensures i already put compute aspect
	mappedBuffer = (unsigned char *)clEnqueueMapBuffer(g_clCommandQueue, g_cl_mem_resultImage, CL_TRUE, CL_MAP_READ, 0, g_w * g_h, 0, NULL, NULL, NULL);

	printf("Results from LOCAL version:\n");
	for(unsigned int i=0;i<g_w * g_h; i++)
	{
		printf("%d is out[%d]\n", mappedBuffer[i], i);
	}
	printf("END\n");

	clEnqueueUnmapMemObject(g_clCommandQueue, g_cl_mem_resultImage, mappedBuffer, 0, NULL, NULL);
	clFinish(g_clCommandQueue);

	return status;
}

int cleanupHost()
{
	//cleanup the mallocd buffers
	if(g_clProgramString != NULL)
	{
		free(g_clProgramString);
		g_clProgramString = NULL;
	}

	//note there is no pointer to free here, via aligned_free or regular free for the original image object
	//TODO: clarify with aaron, he was saying otherwise? might be right...think about it

	if(g_img != NULL)
	{
		free(g_img);
		g_img = NULL;
	}
	
	return SUCCESS;

}

